/*
   Copyright (C) 1996-1997 Robert Muchsel

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   Please address all correspondence concerning the software to
   muchsel@acm.org.
*/

#include "config.h"
#include <asm/system.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/firewall.h>
#include <linux/ip.h>
#include "skip_defs.h"
#include "dynamic.h"
#include "memblk.h"
#include "skipcache.h"
#include "queue.h"
#include "interface.h"

/* #define DEBUG_LINUX_QLEN  */
/* #define DEBUG_LINUX_QUEUE */

#ifdef DEBUG_LINUX_QLEN
static int queue_stat = 0;
#endif

/* Puts a packet into the ip address specific queue. We simply use
   a Linux skb queue. */
int queue_enqueuein(struct skipcache *c, void *interface, void *header,
		    void *pkt)
{
  struct _kernel_queue *q = (struct _kernel_queue *) (c->data + c->q.offset);

#ifdef DEBUG_LINUX_QUEUE
  struct iphdr *ip = ((struct sk_buff *) pkt)->ip_hdr;

  printf("queue: enqueue_in srcip=%08x dstip=%08x\n", ip->saddr, ip->daddr);
#endif
#ifdef DEBUG_LINUX_QLEN
  queue_stat++;
#endif
  skb_queue_tail(&q->inq, (struct sk_buff *) pkt);

  return 0;
}

/* Same for the output queue */
int queue_enqueueout(struct skipcache *c, void *interface, void *header,
		     void *pkt)
{
  struct _kernel_queue *q = (struct _kernel_queue *) (c->data + c->q.offset);

#ifdef DEBUG_LINUX_QUEUE
  struct iphdr *ip = ((struct sk_buff *) pkt)->ip_hdr;

  printf("queue: enqueue_out srcip=%08x dstip=%08x\n", ip->saddr, ip->daddr);
#endif
#ifdef DEBUG_LINUX_QLEN
  queue_stat++;
#endif
  skb_queue_tail(&q->outq, (struct sk_buff *) pkt);

  return 0;
}

/* Free the queue for a given IP address */
int queue_free(void *cptr)
{
  struct skipcache *c = (struct skipcache *) cptr;
  struct _kernel_queue *q = (struct _kernel_queue *) (c->data + c->q.offset);
  struct sk_buff *m;

#ifdef DEBUG_LINUX_QUEUE
  printf("queue: queue_free\n");
#endif

  while ((m = skb_dequeue(&q->outq)) != NULL) {
#ifdef DEBUG_LINUX_QLEN
    queue_stat--;
#endif
    kfree_skb(m, FREE_WRITE);
  }

  while ((m = skb_dequeue(&q->inq)) != NULL) {
#ifdef DEBUG_LINUX_QLEN
    queue_stat--;
#endif
    kfree_skb(m, FREE_WRITE);
  }
 
  
#ifdef DEBUG_LINUX_QLEN
  printk("queue_stat: %i\n", queue_stat);
#endif

  return 0;
}

/* Feed all skbs for a given IP address back into processing */
int queue_feed(void *cptr)
{
  struct skipcache *c = (struct skipcache *) cptr;
  struct _kernel_queue *q = (struct _kernel_queue *) (c->data + c->q.offset);
  struct sk_buff *m;

#ifdef DEBUG_LINUX_QUEUE
  printf("queue: queue_feed\n");
#endif

  while ((m = skb_dequeue(&q->inq)) != NULL) {
#ifdef DEBUG_LINUX_QLEN
    queue_stat--;
#endif
    interface_feed_in(m);
  }

  while ((m = skb_dequeue(&q->outq)) != NULL) {
#ifdef DEBUG_LINUX_QLEN
    queue_stat--;
#endif
    interface_feed_out(m);
  }

#ifdef DEBUG_LINUX_QLEN
  printk("queue_stat: %i\n", queue_stat);
#endif

  return 0;
}

/* Get IP addresses for a specific queue */
int queue_getips(struct skipcache *c, u_char *srcip, u_char *dstip)
{
  int result = -1;
  struct _kernel_queue *q = (struct _kernel_queue *) (c->data + c->q.offset);
  struct sk_buff *m;

#ifdef DEBUG_LINUX_QUEUE
  printf("queue: queue_getips");
#endif

  if ((m = skb_peek(&q->inq)) == NULL)
    m = skb_peek(&q->outq);

  if (m != NULL) {
    *((__u32 *) srcip) = m->ip_hdr->saddr;
    *((__u32 *) dstip) = m->ip_hdr->daddr;
    result = 0;
  }
  else {
    *((__u32 *) srcip) = 0;
    *((__u32 *) dstip) = 0;
  }

#ifdef DEBUG_LINUX_QUEUE
  printf(": srcip=%08x dstip=%08x result=%i\n", *((__u32 *) srcip),
         *((__u32 *) dstip), result);
#endif

  return result;
}


/* initialize input/output queues */
int queue_initqueues(struct skipcache *c)
{
  struct _kernel_queue *q = (struct _kernel_queue *) (c->data + c->q.offset);

  skb_queue_head_init(&q->inq);
  skb_queue_head_init(&q->outq);

  return 0;
}

/* init the queue module */
int queue_init(void)
{
  return 0;
}

/* de-init the queue module */
int queue_exit(void)
{
  return 0;
}
